"""General SAX Utilities."""

from __future__ import annotations

import inspect
import re
import warnings
from collections.abc import Callable, Iterable, Iterator
from copy import deepcopy
from functools import partial, wraps
from hashlib import md5
from pathlib import Path
from typing import (
    Any,
    NamedTuple,
    TypeVar,
    Unpack,
    cast,
    get_args,
    get_origin,
    overload,
)

import jax
import jax.numpy as jnp
import numpy as np
import orjson
import pandas as pd
import yaml
from numpy.exceptions import ComplexWarning

import sax

from .s import reciprocal

T = TypeVar("T")

__all__ = [
    "clean_string",
    "copy_settings",
    "flatten_dict",
    "get_settings",
    "grouped_interp",
    "hash_dict",
    "load_netlist",
    "load_recursive_netlist",
    "maybe",
    "merge_dicts",
    "read",
    "reciprocal",  # re-exported here to not break gplugins
    "rename_params",
    "rename_ports",
    "replace_kwargs",
    "same_args_as",
    "unflatten_dict",
    "update_settings",
]


def maybe(
    func: Callable[..., T], /, exc: type[Exception] = Exception
) -> Callable[..., T | None]:
    """Create a safe version of a function that returns None on exceptions.

    Wraps a function to catch specified exceptions and return None instead of
    raising them. This is useful for optional operations or when you want to
    gracefully handle failures.

    Args:
        func: Function to wrap for safe execution.
        exc: Exception type(s) to catch. Defaults to Exception (catches all).

    Returns:
        Wrapped function that returns None when the specified exception occurs.

    Example:
        ```python
        # Safe division that returns None for division by zero
        safe_divide = maybe(lambda x, y: x / y, ZeroDivisionError)
        result = safe_divide(10, 0)  # Returns None instead of raising

        # Safe file reading
        safe_read = maybe(lambda f: open(f).read(), FileNotFoundError)
        content = safe_read("nonexistent.txt")  # Returns None
        ```
    """

    @wraps(func)
    def new_func(*args: Any, **kwargs: Any) -> T | None:  # noqa: ANN401
        try:
            return func(*args, **kwargs)
        except exc:
            return None

    return new_func


def copy_settings(settings: sax.Settings) -> sax.Settings:
    """Create a deep copy of a settings dictionary.

    Creates a deep copy of a settings dictionary to avoid unintended mutations
    of shared parameter dictionaries.

    Args:
        settings: Settings dictionary to copy.

    Returns:
        Deep copy of the input settings dictionary.

    Example:
        ```python
        original = {"wl": 1.55, "temp": 300, "nested": {"param": 1.0}}
        copied = copy_settings(original)
        copied["nested"]["param"] = 2.0  # Doesn't affect original
        ```
    """
    return deepcopy(settings)


def read(content_or_filename: str | Path | sax.IOLike) -> str:
    r"""Read content from string, file path, or file-like object.

    Flexible content reader that can handle string content directly, file paths,
    or file-like objects. Automatically detects the input type and reads accordingly.

    Args:
        content_or_filename: Content as string (if contains newlines), file path,
            or file-like object with read() method.

    Returns:
        Content as string.

    Example:
        ```python
        # Direct string content
        content = read("line1\\nline2")

        # From file path
        content = read("config.yaml")

        # From file-like object
        from io import StringIO

        content = read(StringIO("data"))
        ```
    """
    if isinstance(content_or_filename, str) and "\n" in content_or_filename:
        return content_or_filename

    if isinstance(content_or_filename, str | Path):
        return Path(content_or_filename).read_text()

    return content_or_filename.read()


def load_netlist(content_or_filename: str | Path | sax.IOLike) -> sax.Netlist:
    """Load a SAX netlist from YAML content or file.

    Parses YAML content to create a SAX netlist dictionary. The YAML should
    contain instances, connections, and ports sections.

    Args:
        content_or_filename: YAML content as string, file path, or file-like object.

    Returns:
        Parsed netlist dictionary.

    Example:
        ```python
        # Load from file
        netlist = load_netlist("circuit.yml")

        # Load from YAML string
        yaml_content = '''
        instances:
          wg1:
            component: waveguide
        ports:
          in: wg1,in
          out: wg1,out
        '''
        netlist = load_netlist(yaml_content)
        ```
    """
    return yaml.safe_load(read(content_or_filename))


def load_recursive_netlist(
    top_level_path: str | Path,
    ext: str = ".pic.yml",
) -> sax.RecursiveNetlist:
    """Load a SAX recursive netlist from a directory of YAML files.

    Recursively loads all YAML files with the specified extension from a directory
    to create a recursive netlist. Each file becomes a component in the recursive
    netlist, with the filename (without extension) as the component name.

    Args:
        top_level_path: Path to the top-level netlist file.
        ext: File extension to search for. Defaults to ".pic.yml".

    Returns:
        Recursive netlist dictionary mapping component names to their netlists.

    Example:
        ```python
        # Load all .pic.yml files in directory
        recnet = load_recursive_netlist("circuits/main.pic.yml")
        # Result: {"main": {...}, "component1": {...}, "component2": {...}}
        ```
    """
    top_level_path = Path(top_level_path)
    folder_path = top_level_path.parent

    def _net_name(path: Path) -> sax.Name:
        return clean_string(path.name.removesuffix(ext))

    recnet = {_net_name(top_level_path): load_netlist(top_level_path)}

    for path in folder_path.rglob(ext):
        recnet[_net_name(path)] = load_netlist(path)

    return recnet


def clean_string(
    s: str, dot: str = "p", minus: str = "m", other: str = "_"
) -> sax.Name:
    """Clean a string to create a valid Python identifier.

    Converts arbitrary strings to valid Python identifiers by replacing special
    characters and ensuring the result starts with a letter or underscore.

    Args:
        s: String to clean.
        dot: Replacement for dots. Defaults to "p".
        minus: Replacement for minus/dash. Defaults to "m".
        other: Replacement for other special characters. Defaults to "_".

    Returns:
        Valid Python identifier string.

    Raises:
        ValueError: If cleaning fails to produce a valid identifier.

    Example:
        ```python
        clean_string("my-component.v2")  # Result: "my_componentpv2"
        clean_string("2stage_amp")  # Result: "_2stage_amp"
        clean_string("TE-mode")  # Result: "TEm_mode"
        ```
    """
    s = s.strip()
    s = s.replace(".", dot)  # dot
    s = s.replace("-", minus)  # minus
    s = re.sub("[^0-9a-zA-Z_]", other, s)
    if s[0] in "0123456789":
        s = "_" + s
    if not s.isidentifier():
        msg = f"failed to clean string to a valid python identifier: {s}"
        raise ValueError(msg)
    return s


def get_settings(model: sax.Model | sax.ModelFactory) -> sax.Settings:
    """Extract default parameter settings from a SAX model function.

    Inspects a model function's signature to extract default parameter values.
    This is useful for understanding what parameters a model accepts and their
    default values. Also supports Unpack[TypedDict] annotations on **kwargs.

    Args:
        model: SAX model function or model factory to inspect.

    Returns:
        Dictionary of parameter names and their default values.

    Example:
        ```python
        def my_model(wl=1.55, length=10.0, neff=2.4):
            return some_s_matrix


        settings = get_settings(my_model)
        # Result: {"wl": 1.55, "length": 10.0, "neff": 2.4}
        ```
    """
    signature = inspect.signature(model)
    settings: sax.Settings = {}

    for k, v in signature.parameters.items():
        # Handle regular parameters with defaults
        if v.default is not inspect.Parameter.empty:
            settings[k] = v.default
            continue

        # Handle **kwargs with Unpack[TypedDict]
        if (
            v.kind == inspect.Parameter.VAR_KEYWORD
            and v.annotation is not inspect.Parameter.empty
        ):
            origin = get_origin(v.annotation)
            if origin is Unpack:
                args = get_args(v.annotation)
                if args and len(args) > 0:
                    typed_dict_cls = args[0]
                    # Check if it's a TypedDict by looking for __annotations__
                    if hasattr(typed_dict_cls, "__annotations__"):
                        annotations = typed_dict_cls.__annotations__
                        for field_name in annotations:
                            # Check if there's a default value (Python 3.13+)
                            try:
                                default_value = getattr(typed_dict_cls, field_name)
                                # Verify it's not a method or function
                                if not inspect.ismethod(
                                    default_value
                                ) and not inspect.isfunction(default_value):
                                    settings[field_name] = default_value
                            except AttributeError:
                                # Field has no default value
                                pass

    return cast(sax.Settings, settings)


def merge_dicts(*dicts: dict) -> dict:
    """Merge multiple dictionaries with support for nested merging.

    Recursively merges multiple dictionaries, with later dictionaries
    taking precedence over earlier ones. Nested dictionaries are merged
    recursively rather than replaced entirely.

    Args:
        *dicts: Variable number of dictionaries to merge.

    Returns:
        Merged dictionary with nested structures preserved.

    Example:
        ```python
        dict1 = {"a": 1, "nested": {"x": 10}}
        dict2 = {"b": 2, "nested": {"y": 20}}
        merged = merge_dicts(dict1, dict2)
        # Result: {"a": 1, "b": 2, "nested": {"x": 10, "y": 20}}
        ```
    """
    num_dicts = len(dicts)

    if num_dicts < 1:
        return {}

    if num_dicts == 1:
        return dict(_generate_merged_dict(dicts[0], {}))

    if len(dicts) == 2:
        return dict(_generate_merged_dict(dicts[0], dicts[1]))

    return merge_dicts(dicts[0], merge_dicts(*dicts[1:]))


def replace_kwargs(func: Callable, **kwargs: sax.SettingsValue) -> None:
    """Change the kwargs signature of a function."""
    sig = inspect.signature(func)
    settings = [
        inspect.Parameter(k, inspect.Parameter.KEYWORD_ONLY, default=v)
        for k, v in kwargs.items()
    ]
    func.__signature__ = sig.replace(parameters=settings)  # type: ignore # noqa: PGH003


def update_settings(
    settings: sax.Settings, *compnames: str, **kwargs: sax.SettingsValue
) -> sax.Settings:
    """Update a nested settings dictionary with new parameter values.

    Updates settings for specific components or globally. Supports nested
    parameter dictionaries and selective component updates.

    Args:
        settings: Original settings dictionary to update.
        *compnames: Component names to update. If empty, updates all components.
        **kwargs: Parameter values to update.

    Returns:
        Updated settings dictionary (does not modify original).

    Note:
        - This operation never updates the given settings dictionary in place.
        - Any non-float keyword arguments will be silently ignored.
        - Even though it's possible to update parameter dictionaries in place,
          this function is convenient to apply certain parameters (e.g. wavelength
          'wl' or temperature 'T') globally.

    Example:
        ```python
        settings = {
            "wg1": {"length": 10.0, "neff": 2.4},
            "wg2": {"length": 20.0, "neff": 2.4},
        }
        # Update all components
        updated = update_settings(settings, wl=1.55)

        # Update specific component
        updated = update_settings(settings, "wg1", length=15.0)
        ```
    """
    _settings = {}
    if not compnames:
        for k, v in settings.items():
            if isinstance(v, dict):
                _settings[k] = update_settings(v, **kwargs)
            elif k in kwargs:
                _settings[k] = _try_complex_float(kwargs[k])
            else:
                _settings[k] = _try_complex_float(v)
    else:
        for k, v in settings.items():
            if isinstance(v, dict):
                if k == compnames[0]:
                    _settings[k] = update_settings(v, *compnames[1:], **kwargs)
                else:
                    _settings[k] = v
            else:
                _settings[k] = _try_complex_float(v)
    return _settings


def flatten_dict(dic: dict[str, Any], sep: str = ",") -> dict[str, Any]:
    """Flatten a nested dictionary into a single-level dictionary.

    Converts a nested dictionary into a flat dictionary by concatenating
    nested keys with a separator.

    Args:
        dic: Nested dictionary to flatten.
        sep: Separator to use between nested keys. Defaults to ",".

    Returns:
        Flattened dictionary with concatenated keys.

    Example:
        ```python
        nested = {"a": {"b": 1, "c": {"d": 2}}, "e": 3}
        flat = flatten_dict(nested)
        # Result: {"a,b": 1, "a,c,d": 2, "e": 3}
        ```
    """
    return _flatten_dict(dic, sep=sep)


def unflatten_dict(dic: dict[str, Any], sep: str = ",") -> dict[str, Any]:
    """Unflatten a dictionary by splitting keys and creating nested structure.

    Converts a flattened dictionary back to nested form by splitting keys
    on the separator and creating the nested hierarchy.

    Args:
        dic: Flattened dictionary to unflatten.
        sep: Separator used in the flattened keys. Defaults to ",".

    Returns:
        Nested dictionary with original structure restored.

    Example:
        ```python
        flat = {"a,b": 1, "a,c,d": 2, "e": 3}
        nested = unflatten_dict(flat)
        # Result: {"a": {"b": 1, "c": {"d": 2}}, "e": 3}
        ```
    """
    # from: https://gist.github.com/fmder/494aaa2dd6f8c428cede
    items = {}

    for k, v in dic.items():
        keys = k.split(sep)
        sub_items = items
        for ki in keys[:-1]:
            if ki in sub_items:
                sub_items = sub_items[ki]
            else:
                sub_items[ki] = {}
                sub_items = sub_items[ki]

        sub_items[keys[-1]] = v

    return items


def grouped_interp(
    wl: sax.FloatArray, wls: sax.FloatArray, phis: sax.FloatArray
) -> sax.FloatArray:
    """Perform grouped phase interpolation for optical phase data.

    Grouped interpolation is useful for interpolating phase values where each
    datapoint is doubled (very close together) to give an indication of the
    phase variation at that point. This is common in optical simulations where
    phase unwrapping is needed.

    Args:
        wl: Wavelength points where interpolation is desired.
        wls: Reference wavelength points (1D array).
        phis: Phase values at reference wavelengths (1D array).

    Returns:
        Interpolated phase values at the requested wavelengths.

    Warning:
        This interpolation is only accurate in the range [wls[0], wls[-2])
        (wls[-2] not included). Any extrapolation outside these bounds can
        yield unexpected results!

    Raises:
        ValueError: If wls or phis are not 1D arrays or have mismatched shapes.

    Example:
        ```python
        import jax.numpy as jnp

        # Reference phase data with grouped points
        wls_ref = jnp.array([1.50, 1.501, 1.55, 1.551, 1.60, 1.601])
        phis_ref = jnp.array([0.1, 0.11, 0.5, 0.51, 1.0, 1.01])

        # Interpolate at new wavelengths
        wl_new = jnp.linspace(1.51, 1.59, 10)
        phis_interp = grouped_interp(wl_new, wls_ref, phis_ref)
        ```
    """
    wl = jnp.asarray(wl)
    wls = jnp.asarray(wls)
    phis = jnp.asarray(phis) % (2 * jnp.pi)
    phis = jnp.where(phis > jnp.pi, phis - 2 * jnp.pi, phis)
    if wls.ndim != 1:
        msg = "grouped_interp: wls should be a 1D array"
        raise ValueError(msg)
    if phis.ndim != 1:
        msg = "grouped_interp: wls should be a 1D array"
        raise ValueError(msg)
    if wls.shape != phis.shape:
        msg = "grouped_interp: wls and phis shape does not match"
        raise ValueError(msg)
    return _grouped_interp(wl.reshape(-1), wls, phis).reshape(*wl.shape)


def rename_params(model: sax.Model, renamings: dict[str, str]) -> sax.Model:
    """Rename the parameters of a model function.

    Creates a new model with renamed parameters while preserving the original
    functionality and default values.

    Args:
        model: Model function to rename parameters for.
        renamings: Dictionary mapping old parameter names to new names.

    Returns:
        New model function with renamed parameters.

    Raises:
        ValueError: If multiple old names map to the same new name.

    Example:
        ```python
        def original_model(wavelength=1.55, eff_index=2.4):
            return some_s_matrix


        # Rename parameters to standard names
        renamed_model = rename_params(
            original_model, {"wavelength": "wl", "eff_index": "neff"}
        )
        # Now can call: renamed_model(wl=1.55, neff=2.4)
        ```
    """
    reversed_renamings = {v: k for k, v in renamings.items()}
    if len(reversed_renamings) < len(renamings):
        msg = "Multiple old names point to the same new name!"
        raise ValueError(msg)

    if callable(_model := cast(sax.Model, model)):

        @wraps(_model)
        def new_model(**settings: sax.SettingsValue) -> sax.SType:
            old_settings = {
                reversed_renamings.get(k, k): v for k, v in settings.items()
            }
            return _model(**old_settings)

        new_settings = {renamings.get(k, k): v for k, v in get_settings(_model).items()}
        _replace_kwargs(new_model, **new_settings)

        return cast(sax.Model, new_model)

    msg = "rename_params should be used to decorate a Model."
    raise ValueError(msg)


@overload
def rename_ports(S: sax.SDict, renamings: dict[str, str]) -> sax.SDict: ...


@overload
def rename_ports(S: sax.SCoo, renamings: dict[str, str]) -> sax.SCoo: ...


@overload
def rename_ports(S: sax.SDense, renamings: dict[str, str]) -> sax.SDense: ...


@overload
def rename_ports(S: sax.Model, renamings: dict[str, str]) -> sax.Model: ...


def rename_ports(
    S: sax.SType | sax.Model, renamings: dict[str, str]
) -> sax.SType | sax.Model:
    """Rename the ports of an S-matrix or model.

    Creates a new S-matrix or model with renamed ports while preserving
    all S-parameter values and relationships.

    Args:
        S: S-matrix in any format or model function to rename ports for.
        renamings: Dictionary mapping old port names to new names.

    Returns:
        S-matrix or model with renamed ports.

    Raises:
        ValueError: If the input type is not supported for port renaming.

    Example:
        ```python
        # Rename ports in S-matrix
        s_matrix = {("input", "output"): 0.9}
        renamed_s = rename_ports(s_matrix, {"input": "in", "output": "out"})
        # Result: {("in", "out"): 0.9}


        # Rename ports in model
        def original_model(wl=1.55):
            return {("input", "output"): 0.9}


        renamed_model = rename_ports(original_model, {"input": "in", "output": "out"})
        ```
    """
    if callable(model := S):

        @wraps(model)
        def new_model(**settings: sax.SettingsValue) -> sax.SType:
            return rename_ports(model(**settings), renamings)

        return cast(sax.Model, new_model)

    if isinstance(sdict := S, dict):
        return {(renamings[p1], renamings[p2]): v for (p1, p2), v in sdict.items()}

    if len(scoo := cast(sax.SCoo, S)) == 4:
        Si, Sj, Sx, ports_map = scoo
        ports_map = {renamings[p]: i for p, i in ports_map.items()}
        return Si, Sj, Sx, ports_map

    if len(sdense := cast(sax.SDense, S)) == 2:
        Sx, ports_map = sdense
        ports_map = {renamings[p]: i for p, i in ports_map.items()}
        return Sx, ports_map

    msg = f"Cannot rename ports for type {type(S)}"
    raise ValueError(msg)


def hash_dict(dic: dict) -> int:
    """Compute a hash value for a dictionary.

    Creates a deterministic hash of a dictionary that can contain NumPy arrays
    and nested structures. Useful for caching and change detection.

    Args:
        dic: Dictionary to hash.

    Returns:
        Integer hash value.

    Example:
        ```python
        settings = {"wl": 1.55, "length": 10.0}
        hash_val = hash_dict(settings)
        # Same settings will always produce the same hash
        ```
    """
    return int(
        md5(
            orjson.dumps(
                _numpyfy(dic), option=orjson.OPT_SERIALIZE_NUMPY | orjson.OPT_SORT_KEYS
            )
        ).hexdigest(),
        16,
    )


class Normalization(NamedTuple):
    """Normalization parameters for an array.

    Contains the mean and standard deviation values needed to normalize
    and denormalize arrays. Typically used for machine learning preprocessing.

    Attributes:
        mean: Mean values for normalization.
        std: Standard deviation values for normalization.
    """

    mean: sax.ComplexArray
    std: sax.ComplexArray


def normalization(x: sax.ComplexArray, axis: int | None = None) -> Normalization:
    """Calculate normalization parameters (mean and std) for an array.

    Computes the mean and standard deviation of an array along the specified
    axis for use in normalization operations.

    Args:
        x: Input array to compute normalization parameters for.
        axis: Axis along which to compute statistics. If None, computes over
            the entire array.

    Returns:
        Normalization object containing mean and standard deviation.

    Example:
        ```python
        import jax.numpy as jnp

        data = jnp.array([[1, 2, 3], [4, 5, 6]])
        norm_params = normalization(data, axis=0)
        # Computes mean and std along axis 0
        ```
    """
    if axis is None:
        return Normalization(x.mean(), x.std())
    return Normalization(x.mean(axis), x.std(axis))


def cartesian_product(*arrays: sax.ComplexArray) -> sax.ComplexArray:
    """Calculate the n-dimensional Cartesian product of input arrays.

    Creates all possible combinations of elements from the input arrays,
    useful for parameter sweeps and grid generation.

    Args:
        *arrays: Variable number of arrays to compute Cartesian product for.

    Returns:
        Array where each row is a unique combination of input elements.

    Example:
        ```python
        import jax.numpy as jnp

        x = jnp.array([1, 2])
        y = jnp.array([10, 20])
        product = cartesian_product(x, y)
        # Result: [[1, 10], [1, 20], [2, 10], [2, 20]]
        ```
    """
    ixarrays = jnp.ix_(*arrays)
    barrays = jnp.broadcast_arrays(*ixarrays)
    sarrays = jnp.stack(barrays, -1)
    product = sarrays.reshape(-1, sarrays.shape[-1])
    return product


def normalize(x: sax.ComplexArray, normalization: Normalization) -> sax.ComplexArray:
    """Normalize an array using provided normalization parameters.

    Applies z-score normalization: (x - mean) / std.

    Args:
        x: Array to normalize.
        normalization: Normalization parameters containing mean and std.

    Returns:
        Normalized array with zero mean and unit standard deviation.

    Example:
        ```python
        data = jnp.array([1, 2, 3, 4, 5])
        norm_params = normalization(data)
        normalized = normalize(data, norm_params)
        ```
    """
    mean, std = normalization
    return (x - mean) / std


def denormalize(x: sax.ComplexArray, normalization: Normalization) -> sax.ComplexArray:
    """Denormalize an array using provided normalization parameters.

    Reverses z-score normalization: x * std + mean.

    Args:
        x: Normalized array to denormalize.
        normalization: Normalization parameters containing mean and std.

    Returns:
        Denormalized array with original scale and offset.

    Example:
        ```python
        normalized_data = jnp.array([-1, 0, 1])  # normalized
        norm_params = Normalization(mean=3.0, std=2.0)
        original = denormalize(normalized_data, norm_params)
        # Result: [1, 3, 5]
        ```
    """
    mean, std = normalization
    return x * std + mean


def wide_to_tidy(df: pd.DataFrame, ports: Iterable[str] | None = None) -> pd.DataFrame:
    """Convert a wide-format S-parameter DataFrame to tidy format.

    Args:
        df: Wide-format DataFrame with columns for wavelength, real, and imaginary
            parts of S-parameters.
        ports: Optional list of port names. If not provided, ports will be
            automatically generated as "o1", "o2", etc.

    Returns:
        Tidy-format DataFrame with columns:
            - 'wl': Wavelength values.
            - 'port_in': Input port names.
            - 'port_out': Output port names.
            - 'mode_in': Input mode labels.
            - 'mode_out': Output mode labels.
            - 're': Real part of S-parameters.
            - 'im': Imaginary part of S-parameters.
    """
    df.columns = [re.sub("[^a-z0-9]", "", c.lower()) for c in df.columns]
    df_im = df[["wl", *[c for c in df.columns if "im" in c]]]
    df_im = df_im.melt(id_vars="wl", value_name="im")
    if df_im.shape[0] == 0:
        msg = (
            "Given wide dataframe columns does not seem have columns "
            "with a 'im' (imaginary) marker."
        )
        raise ValueError(msg)
    df = cast(pd.DataFrame, df[["wl", *[c for c in df.columns if "re" in c]]])
    df = df.melt(id_vars="wl", value_name="re")
    if df.shape[0] == 0:
        msg = (
            "Given wide dataframe columns does not seem have columns "
            "with a 're' (real) marker."
        )
        raise ValueError(msg)
    port_idxs = [_get_port_idx_combos(s) for s in df["variable"]]
    port_idxs = sorted({idx for idxs in port_idxs for idx in idxs})
    if not ports:
        ports = [f"o{i + 1}" for i in port_idxs]
    else:
        ports = list(ports)
        for idx in port_idxs:
            try:
                ports[idx]
            except KeyError as e:
                msg = (
                    f"Cannot map port index {idx} to a port name. "
                    "is the given list of ports long enough?"
                )
                raise ValueError(msg) from e
    port_combos = [_get_port_combos(s, ports) for s in df["variable"]]
    df["port_in"] = [_get_port(p) for p, _ in port_combos]
    df["mode_in"] = [_get_mode(p) for p, _ in port_combos]
    df["port_out"] = [_get_port(q) for _, q in port_combos]
    df["mode_out"] = [_get_mode(q) for _, q in port_combos]
    df["im"] = df_im["im"]
    df = cast(
        pd.DataFrame,
        df[["wl", "port_in", "port_out", "mode_in", "mode_out", "re", "im"]],
    )
    return df


def same_args_as(
    func: Callable[..., T],
) -> Callable[[Callable[..., T]], Callable[..., T]]:
    """Have the same signature as the wrapped function."""

    def decorator(wrapped_func: Callable[..., T]) -> Callable[..., T]:
        name = wrapped_func.__name__
        qualname = wrapped_func.__qualname__
        module = wrapped_func.__module__
        doc = wrapped_func.__doc__
        new_func = wraps(func)(wrapped_func)
        new_func.__name__ = name
        new_func.__qualname__ = qualname
        new_func.__module__ = module
        new_func.__doc__ = doc
        return new_func

    return decorator


def _generate_merged_dict(dict1: dict, dict2: dict) -> Iterator[tuple[Any, Any]]:
    # inspired by https://stackoverflow.com/questions/7204805/how-to-merge-dictionaries-of-dictionaries
    keys = {
        **dict.fromkeys(dict1),
        **dict.fromkeys(dict2),
    }  # keep key order, values irrelevant
    for k in keys:
        if k in dict1 and k in dict2:
            v1, v2 = dict1[k], dict2[k]
            if isinstance(v1, dict) or isinstance(v2, dict):
                # note: we used to have only merging when both where dicts,
                # but this way is more flexible - if one is not a dict, we
                # just replace it with an empty dict and continue merging.
                v = dict(
                    _generate_merged_dict(
                        {} if not isinstance(v1, dict) else v1,
                        {} if not isinstance(v2, dict) else v2,
                    )
                )
            else:
                # If neither of the values is a dict, you can't continue merging it.
                # Value from second dict overrides one in first and we move on.
                v = v2
        elif k in dict1:
            v = dict1[k]
        else:  # k in dict2:
            v = dict2[k]

        if isinstance(v, dict):
            yield (k, {**v})  # shallow copy of dict
        else:
            yield (k, v)


def _try_complex_float(f: Any) -> Any:  # noqa: ANN401
    """Try converting an object to float, return unchanged object on fail."""
    with warnings.catch_warnings():
        warnings.filterwarnings(action="error", category=ComplexWarning)
        try:
            return jnp.asarray(f, dtype=float)
        except ComplexWarning:
            return jnp.asarray(f, dtype=complex)
        except (ValueError, TypeError):
            pass

    return f


def _flatten_dict(
    dic: dict[str, Any], sep: str = ",", *, frozen: bool = False, parent_key: str = ""
) -> dict[str, Any]:
    items = []
    for k, v in dic.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(
                _flatten_dict(v, sep=sep, frozen=frozen, parent_key=new_key).items()
            )
        else:
            items.append((new_key, v))

    return dict(items)


@partial(jax.vmap, in_axes=(0, None, None), out_axes=0)
@jax.jit
def _grouped_interp(
    wl: float,  # 0D array (not-vmapped) ; 1D array (vmapped)
    wls: sax.FloatArray,  # 1D array
    phis: sax.FloatArray,  # 1D array
) -> sax.FloatArray:
    dphi_dwl = (phis[1::2] - phis[::2]) / (wls[1::2] - wls[::2])
    phis = phis[::2]
    wls = wls[::2]
    dwl = (wls[1:] - wls[:-1]).mean(0, keepdims=True)

    t = (wl - wls + 1e-5 * dwl) / dwl  # small offset to ensure no values are zero
    t = jnp.where(jnp.abs(t) < 1, t, 0)
    m0 = jnp.where(t > 0, size=1)[0]
    m1 = jnp.where(t < 0, size=1)[0]
    t = t[m0]
    wl0 = wls[m0]
    wl1 = wls[m1]
    phi0 = phis[m0]
    phi1 = phis[m1]
    dphi_dwl0 = dphi_dwl[m0]
    dphi_dwl1 = dphi_dwl[m1]
    _phi0 = phi0 - 0.5 * (wl1 - wl0) * (dphi_dwl0 * (t**2 - 2 * t) - dphi_dwl1 * t**2)
    _phi1 = phi1 - 0.5 * (wl1 - wl0) * (
        dphi_dwl0 * (t - 1) ** 2 - dphi_dwl1 * (t**2 - 1)
    )
    phis = jnp.arctan2(
        (1 - t) * jnp.sin(_phi0) + t * jnp.sin(_phi1),
        (1 - t) * jnp.cos(_phi0) + t * jnp.cos(_phi1),
    )
    return phis


def _replace_kwargs(func: Callable, **kwargs: Any) -> None:  # noqa: ANN401
    """Change the kwargs signature of a function."""
    sig = inspect.signature(func)
    settings = [
        inspect.Parameter(k, inspect.Parameter.KEYWORD_ONLY, default=v)
        for k, v in kwargs.items()
    ]
    func.__signature__ = sig.replace(parameters=settings)  # type: ignore[reportFunctionMemberAccess]


def _numpyfy(obj: Any) -> Any:  # noqa: ANN401
    if not isinstance(obj, dict):
        return np.asarray(obj)
    return {k: _numpyfy(v) for k, v in obj.items()}


def _get_port_idx_combos(s: str) -> tuple[int, int]:
    match1, match2, *_ = list(re.finditer("[0-9]", s))
    idx1 = int(match1.group()) - 1
    idx2 = int(match2.group()) - 1
    return idx1, idx2


def _get_port_combos(s: str, ports: list[str]) -> tuple[str, str]:
    i, j = _get_port_idx_combos(s)
    return ports[i], ports[j]


def _get_port(pm: str) -> str:
    return pm.split("@")[0]


def _get_mode(pm: str) -> str:
    _, *m = pm.split("@")
    return "".join(m) or "1"



"""Sax generic RF models."""

from functools import partial

import jax
import jax.numpy as jnp

import sax


@partial(jax.jit, static_argnames=("n_ports",))
def gamma_0_load(
    f: sax.FloatArrayLike = 5e9,
    gamma_0: sax.ComplexLike = 0,
    n_ports: sax.IntLike = 1,
) -> sax.SDict:
    r"""Connection with given reflection coefficient.

    Args:
        f: Array of frequency points in Hz
        gamma_0: Reflection coefficient Γ₀ of connection
        n_ports: Number of ports in component. The diagonal ports of the matrix
            are set to Γ₀ and the off-diagonal ports to 0.

    Returns:
        S-dictionary where :math:`S = \Gamma_0I_\text{n\_ports}`

    Examples:
        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        f = np.linspace(1e9, 10e9, 500)
        gamma_0 = 0.5 * np.exp(1j * np.pi / 4)
        s = sax.models.rf.gamma_0_load(f=f, gamma_0=gamma_0, n_ports=2)
        plt.figure()
        plt.plot(f / 1e9, np.abs(s[("o1", "o1")]), label="|S11|")
        plt.plot(f / 1e9, np.abs(s[("o2", "o2")]), label="|S22|")
        plt.plot(f / 1e9, np.abs(s[("o1", "o2")]), label="|S12|")
        plt.plot(f / 1e9, np.abs(s[("o2", "o1")]), label="|S21|")
        plt.xlabel("Frequency [GHz]")
        plt.ylabel("Magnitude")
        plt.legend()
        ```
    """
    f = jnp.asarray(f)
    sdict = {
        (f"o{i}", f"o{i}"): jnp.full(len(f), gamma_0) for i in range(1, n_ports + 1)
    }
    sdict |= {
        (f"o{i}", f"o{j}"): jnp.zeros(len(f), dtype=complex)
        for i in range(1, n_ports + 1)
        for j in range(i + 1, n_ports + 1)
    }
    return sax.reciprocal(sdict)


@jax.jit
def tee(f: sax.FloatArrayLike = 5e9) -> sax.SDict:
    """Ideal three-port RF power divider/combiner (T-junction).

    Args:
        f: Array of frequency points in Hz

    Returns:
        S-dictionary representing ideal RF T-junction behavior

    Examples:
        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        f = np.linspace(1e9, 10e9, 500)
        s = sax.models.rf.tee(f=f)
        plt.figure()
        plt.plot(f / 1e9, np.abs(s[("o1", "o2")]) ** 2, label="|S12|^2")
        plt.plot(f / 1e9, np.abs(s[("o1", "o3")]) ** 2, label="|S13|^2")
        plt.plot(f / 1e9, np.abs(s[("o2", "o3")]) ** 2, label="|S23|^2")
        plt.xlabel("Frequency [GHz]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    f = jnp.asarray(f)
    sdict = {(f"o{i}", f"o{i}"): jnp.full(len(f), -1 / 3) for i in range(1, 4)}
    sdict |= {
        (f"o{i}", f"o{j}"): jnp.full(len(f), 2 / 3)
        for i in range(1, 4)
        for j in range(i + 1, 4)
    }
    return sax.reciprocal(sdict)


@jax.jit
def impedance(
    f: sax.FloatArrayLike, z: sax.ComplexLike = 50, z0: sax.ComplexLike = 50
) -> sax.SDict:
    r"""Generalized two-port impedance element.

    Args:
        f: Frequency in Hz
        z: Impedance in Ω
        z0: Reference impedance in Ω.

    Returns:
        S-dictionary representing the impedance element

    References:
        [@pozar2012]

    Examples:
        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        f = np.linspace(1e9, 10e9, 500)
        s = sax.models.rf.impedance(f=f, z=75, z0=50)
        plt.figure()
        plt.plot(f / 1e9, np.abs(s[("o1", "o1")]), label="|S11|")
        plt.plot(f / 1e9, np.abs(s[("o1", "o2")]), label="|S12|")
        plt.plot(f / 1e9, np.abs(s[("o2", "o2")]), label="|S22|")
        plt.xlabel("Frequency [GHz]")
        plt.ylabel("Magnitude")
        plt.legend()
        ```
    """
    f = jnp.asarray(f)
    sdict = {
        ("o1", "o1"): jnp.full(f.shape, z / (z + 2 * z0)),
        ("o1", "o2"): jnp.full(f.shape, 2 * z0 / (2 * z0 + z)),
        ("o2", "o2"): jnp.full(f.shape, z / (z + 2 * z0)),
    }
    return sax.reciprocal(sdict)


@jax.jit
def admittance(f: sax.FloatArrayLike, y: sax.ComplexLike = 1 / 50) -> sax.SDict:
    r"""Generalized two-port admittance element.

    Args:
        f: Frequency in Hz
        y: Admittance in siemens

    Returns:
        S-dictionary representing the admittance element

    References:
        [@pozar2012]

    Examples:
        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        f = np.linspace(1e9, 10e9, 500)
        s = sax.models.rf.admittance(f=f, y=1 / 75)
        plt.figure()
        plt.plot(f / 1e9, np.abs(s[("o1", "o1")]), label="|S11|")
        plt.plot(f / 1e9, np.abs(s[("o1", "o2")]), label="|S12|")
        plt.plot(f / 1e9, np.abs(s[("o2", "o2")]), label="|S22|")
        plt.xlabel("Frequency [GHz]")
        plt.ylabel("Magnitude")
        plt.legend()
        ```
    """
    f = jnp.asarray(f)
    sdict = {
        ("o1", "o1"): jnp.full(f.shape, 1 / (1 + y)),
        ("o1", "o2"): jnp.full(f.shape, y / (1 + y)),
        ("o2", "o2"): jnp.full(f.shape, 1 / (1 + y)),
    }
    return sax.reciprocal(sdict)


@partial(jax.jit, static_argnames=("capacitance",))
def capacitor(
    f: sax.FloatArrayLike = 5e9,
    capacitance: sax.FloatLike = 1e-15,
    z0: sax.ComplexLike = 50,
) -> sax.SDict:
    r"""Ideal two-port capacitor model.

    Args:
        f: Frequency in Hz
        capacitance: Capacitance in Farads
        z0: Reference impedance in Ω.

    Returns:
        S-dictionary representing the capacitor element

    References:
        [@pozar2012]

    Examples:
        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        f = np.linspace(1e9, 10e9, 500)
        s = sax.models.rf.capacitor(f=f, capacitance=1e-12, z0=50)
        plt.figure()
        plt.plot(f / 1e9, np.abs(s[("o1", "o1")]), label="|S11|")
        plt.plot(f / 1e9, np.abs(s[("o1", "o2")]), label="|S12|")
        plt.plot(f / 1e9, np.abs(s[("o2", "o2")]), label="|S22|")
        plt.xlabel("Frequency [GHz]")
        plt.ylabel("Magnitude")
        plt.legend()
        ```
    """
    angular_frequency = 2 * jnp.pi * jnp.asarray(f)
    capacitor_impedance = 1 / (1j * angular_frequency * capacitance)
    return impedance(f=f, z=capacitor_impedance, z0=z0)


@partial(jax.jit, static_argnames=("inductance",))
def inductor(
    f: sax.FloatArrayLike = 5e9,
    inductance: sax.FloatLike = 1e-12,
    z0: sax.ComplexLike = 50,
) -> sax.SDict:
    r"""Ideal two-port inductor model.

    Args:
        f: Frequency in Hz
        inductance: Inductance in Henries
        z0: Reference impedance in Ω.

    Returns:
        S-dictionary representing the inductor element

    References:
        [@pozar2012]

    Examples:
        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        f = np.linspace(1e9, 10e9, 500)
        s = sax.models.rf.inductor(f=f, inductance=1e-9, z0=50)
        plt.figure()
        plt.plot(f / 1e9, np.abs(s[("o1", "o1")]), label="|S11|")
        plt.plot(f / 1e9, np.abs(s[("o1", "o2")]), label="|S12|")
        plt.plot(f / 1e9, np.abs(s[("o2", "o2")]), label="|S22|")
        plt.xlabel("Frequency [GHz]")
        plt.ylabel("Magnitude")
        plt.legend()
        ```
    """
    angular_frequency = 2 * jnp.pi * jnp.asarray(f)
    inductor_impedance = 1j * angular_frequency * inductance
    return impedance(f=f, z=inductor_impedance, z0=z0)


@partial(jax.jit, inline=True, static_argnames=("n_ports"))
def electrical_short(
    f: sax.FloatArrayLike = 5e9,
    n_ports: int = 1,
) -> sax.SDict:
    r"""Electrical short connection Sax model.

    Args:
        f: Array of frequency points in Hz
        n_ports: Number of ports to set as shorted

    Returns:
        S-dictionary where :math:`S = -I_\text{n\_ports}`

    References:
        [@pozar2012]
    """
    return gamma_0_load(f=f, gamma_0=-1, n_ports=n_ports)


@partial(jax.jit, inline=True, static_argnames=("n_ports"))
def electrical_open(
    f: sax.FloatArrayLike = 5e9,
    n_ports: int = 1,
) -> sax.SDict:
    r"""Electrical open connection Sax model.

    Useful for specifying some ports to remain open while not exposing
    them for connections in circuits.

    Args:
        f: Array of frequency points in Hz
        n_ports: Number of ports to set as opened

    Returns:
        S-dictionary where :math:`S = I_\text{n\_ports}`

    References:
        [@pozar2012]
    """
    return gamma_0_load(f=f, gamma_0=1, n_ports=n_ports)



from itertools import product

import jax.numpy as jnp
import pytest

from sax.models import rf


class TestRFModels:
    """Test suite for RF model components."""

    @pytest.fixture
    def freq_array(self) -> jnp.ndarray:
        """Frequency array fixture."""
        return jnp.linspace(1e9, 10e9, 10)

    @pytest.fixture
    def freq_single(self) -> float:
        """Single frequency fixture."""
        return 1e9

    @staticmethod
    def _assert_s_params_dict(s: dict, expected_shape: tuple | None = None) -> None:
        """Helper method to assert common S-parameter dictionary properties."""
        assert isinstance(s, dict)
        if expected_shape is not None:
            assert s[("o1", "o1")].shape == expected_shape

    @staticmethod
    def _assert_s_param(s: dict, port_pair: tuple, expected_value: complex) -> None:
        """Helper method to assert a specific S-parameter value."""
        assert jnp.allclose(s[port_pair], expected_value)

    def test_gamma_0_load(self, freq_array: jnp.ndarray) -> None:
        """Test gamma_0_load with frequency array."""
        s = rf.gamma_0_load(f=freq_array, gamma_0=0.5, n_ports=2)

        self._assert_s_params_dict(s, expected_shape=(len(freq_array),))
        self._assert_s_param(s, ("o1", "o1"), 0.5)
        self._assert_s_param(s, ("o1", "o2"), 0)

    def test_tee(self, freq_array: jnp.ndarray) -> None:
        """Test tee splitter."""
        s = rf.tee(f=freq_array)

        self._assert_s_params_dict(s, expected_shape=(len(freq_array),))
        self._assert_s_param(s, ("o1", "o1"), -1 / 3)
        self._assert_s_param(s, ("o1", "o2"), 2 / 3)

    def test_impedance(self, freq_single: float) -> None:
        """Test impedance element."""
        s = rf.impedance(f=freq_single, z=75, z0=50)

        self._assert_s_params_dict(s, expected_shape=())
        self._assert_s_param(s, ("o1", "o1"), 75 / (75 + 100))

    def test_admittance(self, freq_single: float) -> None:
        """Test admittance element."""
        s = rf.admittance(f=freq_single, y=1 / 75)

        self._assert_s_params_dict(s, expected_shape=())
        self._assert_s_param(s, ("o1", "o1"), 1 / (1 + 1 / 75))

    def test_capacitor(self, freq_single: float) -> None:
        """Test capacitor element."""
        s = rf.capacitor(f=freq_single, capacitance=1e-12, z0=50)

        self._assert_s_params_dict(s, expected_shape=())

        angular_frequency = 2 * jnp.pi * freq_single
        capacitor_impedance = 1 / (1j * angular_frequency * 1e-12)
        expected_s11 = capacitor_impedance / (capacitor_impedance + 100)
        self._assert_s_param(s, ("o1", "o1"), expected_s11)

    def test_inductor(self, freq_single: float) -> None:
        """Test inductor element."""
        s = rf.inductor(f=freq_single, inductance=1e-9, z0=50)

        self._assert_s_params_dict(s)

        angular_frequency = 2 * jnp.pi * freq_single
        inductor_impedance = 1j * angular_frequency * 1e-9
        expected_s11 = inductor_impedance / (inductor_impedance + 100)
        self._assert_s_param(s, ("o1", "o1"), expected_s11)

    @pytest.mark.parametrize("n_ports", [1, 2, 3])
    def test_electrical_short(self, freq_array: jnp.ndarray, n_ports: int) -> None:
        """Test electrical_short with frequency array."""
        s = rf.electrical_short(f=freq_array, n_ports=n_ports)

        self._assert_s_params_dict(s, expected_shape=(len(freq_array),))
        for i, j in product(range(1, n_ports + 1), repeat=2):
            expected_value = -1 if i == j else 0
            self._assert_s_param(s, (f"o{i}", f"o{j}"), expected_value)

    @pytest.mark.parametrize("n_ports", [1, 2, 3])
    def test_electrical_open(self, freq_array: jnp.ndarray, n_ports: int) -> None:
        """Test electrical_open with frequency array."""
        s = rf.electrical_open(f=freq_array, n_ports=n_ports)

        self._assert_s_params_dict(s, expected_shape=(len(freq_array),))
        for i, j in product(range(1, n_ports + 1), repeat=2):
            expected_value = 1 if i == j else 0
            self._assert_s_param(s, (f"o{i}", f"o{j}"), expected_value)



"""SAX Coupler Models."""

from __future__ import annotations

import jax
import jax.numpy as jnp
from pydantic import validate_call

import sax


@jax.jit
@validate_call
def coupler_ideal(
    *,
    wl: sax.FloatArrayLike = sax.WL_C,
    coupling: sax.FloatArrayLike = 0.5,
) -> sax.SDict:
    r"""Ideal 2x2 directional coupler model.

    ```{svgbob}
     in1          out1
      o2          o3
       *          *
        \        /
         '------'
         coupling
         .------.
        /        \
       *          *
      o1          o4
     in0          out0
    ```

    Args:
        wl: the wavelength of the simulation in micrometers.
        coupling: Power coupling coefficient between 0 and 1. Defaults to 0.5.

    Returns:
        The coupler s-matrix

    Examples:
        Ideal coupler:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.coupler_ideal(
            wl=wl,
            coupling=0.3,
        )
        thru = np.abs(s[("o1", "o4")]) ** 2
        cross = np.abs(s[("o1", "o3")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.plot(wl, cross, label="cross")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    one = jnp.ones_like(wl)
    kappa = jnp.asarray(coupling**0.5) * one
    tau = jnp.asarray((1 - coupling) ** 0.5) * one
    p = sax.PortNamer(2, 2)
    return sax.reciprocal(
        {
            (p.in0, p.out0): tau,
            (p.in0, p.out1): 1j * kappa,
            (p.in1, p.out0): 1j * kappa,
            (p.in1, p.out1): tau,
        },
    )


@jax.jit
@validate_call
def coupler(
    *,
    wl: sax.FloatArrayLike = sax.WL_C,
    wl0: sax.FloatArrayLike = sax.WL_C,
    length: sax.FloatArrayLike = 0.0,
    coupling0: sax.FloatArrayLike = 0.2,
    dk1: sax.FloatArrayLike = 1.2435,
    dk2: sax.FloatArrayLike = 5.3022,
    dn: sax.FloatArrayLike = 0.02,
    dn1: sax.FloatArrayLike = 0.1169,
    dn2: sax.FloatArrayLike = 0.4821,
) -> sax.SDict:
    r"""Dispersive directional coupler model.

    ```{svgbob}
            in1                out1
             o2                o3
              *                *
               \              /
                '------------'
    coupling0 /2   coupling   coupling0 /2
                .------------.
               /  <-length -> \
              *                *
             o1                o4
            in0                out0
    ```

    Args:
        wl: Operating wavelength in micrometers. Can be a scalar or array for
            multi-wavelength simulations. Defaults to 1.55 μm.
        wl0: Reference wavelength in micrometers for dispersion expansion.
            Typically the center design wavelength. Defaults to 1.55 μm.
        length: Coupling length in micrometers. This is the length over which
            the two waveguides are in close proximity. Defaults to 0.0 μm.
        coupling0: Base coupling coefficient at the reference wavelength from
            bend regions or other coupling mechanisms. Obtained from FDTD
            simulations or measurements. Defaults to 0.2.
        dk1: First-order derivative of coupling coefficient with respect to
            wavelength (∂κ/∂λ). Units: μm⁻¹. Defaults to 1.2435.
        dk2: Second-order derivative of coupling coefficient with respect to
            wavelength (∂²κ/∂λ²). Units: μm⁻². Defaults to 5.3022.
        dn: Effective index difference between even and odd supermodes at
            the reference wavelength. Determines beating length. Defaults to 0.02.
        dn1: First-order derivative of effective index difference with respect
            to wavelength (∂Δn/∂λ). Units: μm⁻¹. Defaults to 0.1169.
        dn2: Second-order derivative of effective index difference with respect
            to wavelength (∂²Δn/∂λ²). Units: μm⁻². Defaults to 0.4821.

    Returns:
        The coupler s-matrix

    Examples:
        Basic dispersive coupler:

    ```python
    # mkdocs: render
    import matplotlib.pyplot as plt
    import numpy as np
    import sax

    sax.set_port_naming_strategy("optical")

    wl = sax.wl_c()
    s = sax.models.coupler(
        wl=wl,
        length=15.7,
        coupling0=0.0,
        dn=0.02,
    )
    thru = np.abs(s[("o1", "o4")]) ** 2
    cross = np.abs(s[("o1", "o3")]) ** 2
    plt.figure()
    plt.plot(wl, thru, label="thru")
    plt.plot(wl, cross, label="cross")
    plt.xlabel("Wavelength [μm]")
    plt.ylabel("Power")
    plt.legend()
    ```

    Note:
        The coupling strength follows the formula:

        κ_total = κ₀ + κ₁ + κ_length

        Where:

        - κ₀ = coupling0 + dk1*(λ-λ₀) + 0.5*dk2*(λ-λ₀)²
        - κ₁ = π*Δn(λ)/λ
        - κ_length represents the distributed coupling over the interaction length

    """
    dwl = wl - wl0
    dn = dn + dn1 * dwl + 0.5 * dn2 * dwl**2
    kappa0 = coupling0 + dk1 * dwl + 0.5 * dk2 * dwl**2
    kappa1 = jnp.pi * dn / wl

    tau = jnp.cos(kappa0 + kappa1 * length)
    kappa = -jnp.sin(kappa0 + kappa1 * length)
    p = sax.PortNamer(2, 2)
    return sax.reciprocal(
        {
            (p.in0, p.out0): tau,
            (p.in0, p.out1): 1j * kappa,
            (p.in1, p.out0): 1j * kappa,
            (p.in1, p.out1): tau,
        }
    )


@jax.jit
@validate_call
def grating_coupler(
    *,
    wl: sax.FloatArrayLike = sax.WL_C,
    wl0: sax.FloatArrayLike = sax.WL_C,
    loss: sax.FloatArrayLike = 0.0,
    reflection: sax.FloatArrayLike = 0.0,
    reflection_fiber: sax.FloatArrayLike = 0.0,
    bandwidth: sax.FloatArrayLike = 40e-3,
) -> sax.SDict:
    """Grating coupler model for fiber-chip coupling.

    ```{svgbob}
                  out0
                  o2
             /  /  /  /
            /  /  /  /  fiber
           /  /  /  /
           _   _   _
     o1  _| |_| |_| |__
    in0
    ```

    Args:
        wl: Operating wavelength in micrometers. Can be a scalar or array for
            spectral analysis. Defaults to 1.55 μm.
        wl0: Center wavelength in micrometers where peak transmission occurs.
            This is the design wavelength of the grating. Defaults to 1.55 μm.
        loss: Insertion loss in dB at the center wavelength. Includes coupling
            efficiency losses, scattering, and mode mismatch. Defaults to 0.0 dB.
        reflection: Reflection coefficient from the waveguide side (chip interface).
            Represents reflections back into the waveguide from grating discontinuities.
            Range: 0 to 1. Defaults to 0.0.
        reflection_fiber: Reflection coefficient from the fiber side (top interface).
            Represents reflections back toward the fiber from the grating surface.
            Range: 0 to 1. Defaults to 0.0.
        bandwidth: 3dB bandwidth in micrometers. Determines the spectral width
            of the Gaussian transmission profile. Typical values: 20-50 nm.
            Defaults to 40e-3 μm (40 nm).

    Returns:
        The grating coupler s-matrix

    Examples:
        Basic grating coupler:

    ```python
    # mkdocs: render
    import matplotlib.pyplot as plt
    import numpy as np
    import sax

    sax.set_port_naming_strategy("optical")

    wl = np.linspace(1.5, 1.6, 101)
    s = sax.models.grating_coupler(
        wl=wl,
        loss=3.0,
        bandwidth=0.035,
    )
    plt.figure()
    plt.plot(wl, np.abs(s[("o1", "o2")]) ** 2, label="Transmission")
    plt.xlabel("Wavelength (μm)")
    plt.ylabel("Power")
    plt.legend()
    ```

    Note:
        The transmission power profile follows a Gaussian shape:
        P(λ) = P₀ * exp(-((λ-λ₀)/σ)²)

        The amplitude transmission is:
        A(λ) = A₀ * exp(-((λ-λ₀)/σ)²/2)

        Where σ = bandwidth / (2*√(2*ln(2))) converts FWHM to Gaussian width.

        This model assumes:

        - Gaussian spectral response (typical for uniform gratings)
        - Wavelength-independent loss coefficient
        - Linear polarization (TE or TM)
        - Single-mode fiber coupling
        - No higher-order diffraction effects

    """
    one = jnp.ones_like(wl)
    reflection = jnp.asarray(reflection) * one
    reflection_fiber = jnp.asarray(reflection_fiber) * one
    amplitude = jnp.asarray(10 ** (-loss / 20)) * one
    wl0_array = jnp.asarray(wl0) * one
    sigma = jnp.asarray(bandwidth / (2 * jnp.sqrt(2 * jnp.log(2)))) * one
    transmission = amplitude * jnp.exp(-((wl - wl0_array) ** 2) / (4 * sigma**2))
    p = sax.PortNamer(1, 1)
    return sax.reciprocal(
        {
            (p.in0, p.in0): reflection,
            (p.in0, p.out0): transmission,
            (p.out0, p.in0): transmission,
            (p.out0, p.out0): reflection_fiber,
        }
    )



"""SAX Default Models."""

from __future__ import annotations

import jax
import jax.numpy as jnp
from pydantic import validate_call

import sax


@jax.jit
@validate_call
def straight(
    *,
    wl: sax.FloatArrayLike = sax.WL_C,
    wl0: sax.FloatArrayLike = sax.WL_C,
    neff: sax.FloatArrayLike = 2.34,
    ng: sax.FloatArrayLike = 3.4,
    length: sax.FloatArrayLike = 10.0,
    loss_dB_cm: sax.FloatArrayLike = 0.0,
) -> sax.SDict:
    """Dispersive straight waveguide model.

    ```{svgbob}
    in0             out0
     o1 =========== o2
    ```

    Args:
        wl: The wavelength in micrometers.
        wl0: The center wavelength used for dispersion calculation.
        neff: The Effective refractive index at the center wavelength.
        ng: The Group refractive index at the center wavelength.
        length: The length of the waveguide in micrometers.
        loss_dB_cm: The Propagation loss in dB/cm.

    Returns:
        S-matrix dictionary containing the complex transmission coefficient.

    Examples:
        Lossless waveguide:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.straight(wl=wl, coupling=0.3)
        thru = np.abs(s[("o1", "o2")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    dwl: sax.FloatArray = sax.into[sax.FloatArray](wl) - wl0
    dneff_dwl = (ng - neff) / wl0
    _neff = neff - dwl * dneff_dwl
    phase = 2 * jnp.pi * _neff * length / wl
    amplitude = jnp.asarray(10 ** (-1e-4 * loss_dB_cm * length / 20), dtype=complex)
    transmission = amplitude * jnp.exp(1j * phase)
    p = sax.PortNamer(1, 1)
    return sax.reciprocal(
        {
            (p.in0, p.out0): transmission,
        },
    )


@jax.jit
@validate_call
def attenuator(
    *,
    wl: sax.FloatArrayLike = sax.WL_C,
    loss: sax.FloatArrayLike = 0.0,
) -> sax.SDict:
    """Simple optical attenuator model.

    ```{svgbob}
    in0             out0
     o1 =========== o2
    ```

    Args:
        wl: The wavelength in micrometers.
        loss: Attenuation in decibels (dB).

    Returns:
        S-matrix dictionary containing the complex transmission coefficient.

    Examples:
        Attenuator:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.attenuator(wl=wl, loss=3.0)
        thru = np.abs(s[("o1", "o2")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    one = jnp.ones_like(jnp.asarray(wl))
    transmission = jnp.asarray(10 ** (-loss / 20), dtype=complex) * one
    p = sax.PortNamer(1, 1)
    return sax.reciprocal(
        {
            (p.in0, p.out0): transmission,
        }
    )


@jax.jit
@validate_call
def phase_shifter(
    wl: sax.FloatArrayLike = sax.WL_C,
    neff: sax.FloatArrayLike = 2.34,
    voltage: sax.FloatArrayLike = 0,
    length: sax.FloatArrayLike = 10,
    loss: sax.FloatArrayLike = 0.0,
) -> sax.SDict:
    """Simple voltage-controlled phase shifter model.

    ```{svgbob}
    in0             out0
     o1 =========== o2
    ```

    Args:
        wl: The wavelength in micrometers.
        neff: The Effective index of the unperturbed waveguide mode.
        voltage: The Applied voltage in volts. The phase shift is assumed to be
            linearly proportional to voltage with a coefficient of π rad/V.
            Positive voltage increases the phase. Defaults to 0 V.
        length: The length of the phase shifter in micrometers.
        loss: Additional loss in dB introduced by the active region.

    Returns:
        S-matrix dictionary containing the complex-valued transmission coefficient.

    Examples:
        Phase shifter:

        ```python
        # mkdocs: render
        import matplotlib.pyplot as plt
        import numpy as np
        import sax

        sax.set_port_naming_strategy("optical")

        wl = sax.wl_c()
        s = sax.models.phase_shifter(wl=wl, loss=3.0)
        thru = np.abs(s[("o1", "o2")]) ** 2
        plt.figure()
        plt.plot(wl, thru, label="thru")
        plt.xlabel("Wavelength [μm]")
        plt.ylabel("Power")
        plt.legend()
        ```
    """
    deltaphi = voltage * jnp.pi
    phase = 2 * jnp.pi * neff * length / wl + deltaphi
    amplitude = jnp.asarray(10 ** (-loss * length / 20), dtype=complex)
    transmission = amplitude * jnp.exp(1j * phase)
    p = sax.PortNamer(1, 1)
    return sax.reciprocal(
        {
            (p.o1, p.o2): transmission,
        }
    )



"""SAX Types and type coercions.

This module provides the core type system for SAX, including type validation
functions and type aliases for scalars and arrays. It handles type coercion
between Python/NumPy types and JAX arrays with proper validation.

References:
    Numpy type reference: https://numpy.org/doc/stable/reference/arrays.scalars.html
"""

from __future__ import annotations

__all__ = [
    "ArrayLike",
    "Bool",
    "BoolArray",
    "BoolArrayLike",
    "BoolLike",
    "Complex",
    "ComplexArray",
    "ComplexArray1D",
    "ComplexArray1DLike",
    "ComplexArrayLike",
    "ComplexLike",
    "Float",
    "FloatArray",
    "FloatArray1D",
    "FloatArray1DLike",
    "FloatArray2D",
    "FloatArray2DLike",
    "FloatArrayLike",
    "FloatLike",
    "IOLike",
    "Int",
    "IntArray",
    "IntArray1D",
    "IntArray1DLike",
    "IntArrayLike",
    "IntLike",
]

from collections.abc import Callable
from contextlib import suppress
from functools import partial, wraps
from types import UnionType
from typing import (
    Annotated,
    Any,
    Literal,
    LiteralString,
    Protocol,
    TypeAlias,
    TypeVar,
    cast,
    get_args,
    overload,
)

import jax
import jax.numpy as jnp
import numpy as np
from jaxtyping import Array
from pydantic import BeforeValidator, PlainValidator
from pydantic_core import PydanticCustomError

T = TypeVar("T")

ArrayLike: TypeAlias = Array | np.ndarray
"""Anything that can turn into an array with ndim>=1."""


def _val(fun: Callable, **val_kwargs: Any) -> Callable:
    @wraps(fun)
    def new(*args: Any, **kwargs: Any) -> Any:
        kwargs.update(val_kwargs)
        try:
            return fun(*args, **kwargs)
        except TypeError as e:
            msg = str(e)
            if ":" in msg:
                subtype, msg = msg.split(":", 1)
            else:
                subtype = ""
            raise PydanticCustomError(
                cast(LiteralString, subtype),
                cast(LiteralString, msg),
            ) from e

    return new


def val(fun: Callable, **val_kwargs: Any) -> PlainValidator:
    """Create a PlainValidator from a validation function.

    Args:
        fun: The validation function to wrap.
        **val_kwargs: Additional keyword arguments to pass to the validation function.

    Returns:
        A Pydantic PlainValidator wrapping the function.

    Raises:
        PydanticCustomError: If validation fails.
    """
    return PlainValidator(_val(fun, **val_kwargs))


def bval(fun: Callable, **val_kwargs: Any) -> BeforeValidator:
    """Create a BeforeValidator from a validation function.

    Args:
        fun: The validation function to wrap.
        **val_kwargs: Additional keyword arguments to pass to the validation function.

    Returns:
        A Pydantic PlainValidator wrapping the function.

    Raises:
        PydanticCustomError: If validation fails.
    """
    return BeforeValidator(_val(fun, **val_kwargs))


def _val_item_type(
    obj: Any,
    *,
    strict: bool,
    cast: bool,
    type_cast: Callable[..., T],
    type_def: Any,
    type_name: str,
) -> T:
    """Validate and optionally cast a scalar value to a specific type.

    Args:
        obj: The object to validate.
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the target type.
        type_cast: Function to cast to the target type.
        type_def: Type definition for validation.
        type_name: Name of the type for error messages.

    Returns:
        The validated (and optionally cast) value.

    Raises:
        TypeError: If validation fails.
    """
    from ..utils import maybe

    item = _val_0d(obj, type_name=type_name).item()
    if not isinstance(item, _get_annotated_type(type_def)):
        arr = maybe(np.asarray)(item)
        if arr is None or not np.can_cast(arr, type_cast, casting="same_kind"):  # type: ignore[arg-type]
            msg = f"NOT_{type_name.upper()}: Cannot validate {obj!r} into {type_name}."
            raise TypeError(msg)
        if strict:
            msg = (
                f"NOT_{type_name.upper()}: Strict validation does not allow casting "
                f"{obj!r} [dtype={arr.dtype}] into {type_name}. "
                f"Note: use {type_name}Like type for less strict type checking."
            )
            raise TypeError(msg)
    if cast:
        return type_cast(item)
    return item


@overload
def val_bool(obj: Any, *, strict: bool = ..., cast: Literal[True] = True) -> Bool: ...


@overload
def val_bool(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> BoolLike: ...


def val_bool(obj: Any, *, strict: bool = False, cast: bool = True) -> BoolLike:
    """Validate and optionally cast an object to a boolean.

    Args:
        obj: The object to validate.
        strict: Whether to use strict validation (no casting allowed).
        cast: Whether to cast the result to bool.

    Returns:
        The validated boolean value.

    Raises:
        TypeError: If validation fails.

    Examples:
        Validate and cast an object to a boolean:

        ```python
        import sax.saxtypes as sxt

        # Valid boolean values
        result = sxt.val_bool(True)  # True
        result = sxt.val_bool(1)  # True (cast from int)
        result = sxt.val_bool(0.0)  # False (cast from float)
        ```
    """
    return _val_item_type(
        obj,
        strict=strict,
        cast=cast,
        type_cast=bool,
        type_def=Bool,
        type_name="Bool",
    )


Bool: TypeAlias = Annotated[bool | np.bool_, val(val_bool, strict=False)]
"""Any boolean value (Python bool or NumPy boolean)."""


@overload
def val_int(obj: Any, *, strict: bool = ..., cast: Literal[True] = True) -> Int: ...


@overload
def val_int(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> IntLike: ...


def val_int(obj: Any, *, strict: bool = False, cast: bool = True) -> IntLike:
    """Validate and optionally cast an object to an integer.

    Args:
        obj: The object to validate.
        strict: Whether to use strict validation (no casting from bool).
        cast: Whether to cast the result to int.

    Returns:
        The validated integer value.

    Raises:
        TypeError: If validation fails.

    Examples:
        Validate and cast an object to an integer:

        ```python
        import sax.saxtypes as sxt

        # Valid integer values
        result = sxt.val_int(42)  # 42
        result = sxt.val_int(3.0)  # 3 (cast from float)
        result = sxt.val_int("5")  # 5 (cast from string)
        ```
    """
    from ..utils import maybe

    if strict and maybe(partial(val_bool, strict=False, cast=False))(obj) is not None:
        msg = (
            f"NOT_INT: Strict validation does not allow casting {obj} [bool] into Int. "
        )
        raise TypeError(msg)
    return _val_item_type(
        obj,
        strict=strict,
        cast=cast,
        type_cast=int,
        type_def=Int,
        type_name="Int" if strict else "IntLike",
    )


Int: TypeAlias = Annotated[int | np.signedinteger, val(val_int, strict=False)]
"""Any signed integer (Python int or NumPy signed integer)."""


@overload
def val_float(obj: Any, *, strict: bool = ..., cast: Literal[True] = True) -> Float: ...


@overload
def val_float(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> FloatLike: ...


def val_float(obj: Any, *, strict: bool = False, cast: bool = True) -> FloatLike:
    """Validate and optionally cast an object to a float.

    Args:
        obj: The object to validate.
        strict: Whether to use strict validation.
        cast: Whether to cast the result to float.

    Returns:
        The validated float value.

    Raises:
        TypeError: If validation fails.

    Examples:
        Validate and cast an object to a float:

        ```python
        import sax.saxtypes as sxt

        # Valid float values
        result = sxt.val_float(3.14)  # 3.14
        result = sxt.val_float(42)  # 42.0 (cast from int)
        result = sxt.val_float("2.5")  # 2.5 (cast from string)
        ```
    """
    return _val_item_type(
        obj,
        strict=strict,
        cast=cast,
        type_cast=float,
        type_def=Float,
        type_name="Float" if strict else "FloatLike",
    )


Float: TypeAlias = Annotated[float | np.floating, val(val_float, strict=False)]
"""Any floating-point number (Python float or NumPy floating)."""


@overload
def val_complex(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> Complex: ...


@overload
def val_complex(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> ComplexLike: ...


def val_complex(obj: Any, *, strict: bool = False, cast: bool = True) -> ComplexLike:
    """Validate and optionally cast an object to a complex number.

    Args:
        obj: The object to validate.
        strict: Whether to use strict validation.
        cast: Whether to cast the result to complex.

    Returns:
        The validated complex value.

    Raises:
        TypeError: If validation fails.

    Examples:
        Validate and cast an object to a complex number:

        ```python
        import sax.saxtypes as sxt

        # Valid complex values
        result = sxt.val_complex(1 + 2j)  # (1+2j)
        result = sxt.val_complex(3.14)  # (3.14+0j) (cast from float)
        result = sxt.val_complex(42)  # (42+0j) (cast from int)
        ```
    """
    return _val_item_type(
        obj,
        strict=strict,
        cast=cast,
        type_cast=complex,
        type_def=Complex,
        type_name="Complex" if strict else "ComplexLike",
    )


Complex: TypeAlias = Annotated[
    complex | np.complexfloating, val(val_complex, strict=False)
]
"""Any complex number (Python complex or NumPy complex floating)."""


def _val_array_type(  # noqa: C901
    obj: Any,
    *,
    strict: bool,
    cast: bool,
    type_def: Any,
    default_dtype: Any,
    type_name: str,
) -> Array:
    """Validate and optionally cast an object to a JAX array of specific type.

    This is the core array validation function used by all array validators.
    It handles dimensionality checking, dtype validation, and casting.

    Args:
        obj: The object to validate.
        strict: Whether to use strict validation (no numpy arrays, no broadcasting).
        cast: Whether to cast the result to the default dtype.
        type_def: Type definition for validation.
        default_dtype: Default dtype to cast to if casting is enabled.
        type_name: Name of the type for error messages.

    Returns:
        The validated JAX array.

    Raises:
        TypeError: If validation fails.
    """
    if strict:
        if isinstance(obj, np.ndarray):
            msg = (
                f"NOT_{type_name.upper()}: Strict validation does not allow casting "
                f"a numpy array {obj!r} {type(obj)} into a jax-array of "
                f"type {type_name}. Note: use {type_name}Like type for less "
                "strict type checking."
            )
            raise TypeError(msg)
        if not isinstance(obj, Array):
            msg = (
                f"NOT_{type_name.upper()}: Strict validation does not allow casting "
                f"the scalar {obj!r} {type(obj)} into an array type [{type_name}]. "
                f"Note: use {type_name}Like type for less strict type checking."
            )
            raise TypeError(msg)

    short_message = f"Cannot validate {obj!r} into a JAX array."
    arr = _val_array(obj, error_message=short_message)
    ndim = _get_annotated_ndim(type_def)
    if ndim is not None and arr.ndim != ndim:
        if strict:
            msg = (
                f"NOT_{type_name.upper()}: Strict validation does not allow "
                f"broadcasting the {int(arr.ndim)}D array {arr!r} into a {ndim}D "
                f"array. Note: use {type_name}Like type for less strict type checking."
            )
            raise TypeError(msg)
        _ndim = int(arr.ndim)
        while arr.shape[0] == 1:
            arr = arr[0]
        if arr.ndim > ndim:
            msg = (
                f"NOT_{type_name.upper()}: Unable to cast the {_ndim}D "
                f"array {obj!r} into a {ndim}D array. "
            )
            raise TypeError(msg)
        while arr.ndim < ndim:
            arr = arr[None]
    if not np.issubdtype(arr.dtype, _get_annotated_dtype(type_def)):
        if not np.can_cast(arr, default_dtype, casting="same_kind"):
            msg = f"NOT_{type_name.upper()}: Cannot validate {obj!r} into {type_name}."
            raise TypeError(msg)
        if strict:
            msg = (
                f"NOT_{type_name.upper()}: Strict validation does not allow casting "
                f"{obj!r} [dtype={arr.dtype}] into {type_name}. "
                f"Note: use {type_name}Like type for less strict type checking."
            )
            raise TypeError(msg)
    if cast:
        return jnp.asarray(arr, dtype=default_dtype)
    return arr


@overload
def val_bool_array(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> BoolArray: ...


@overload
def val_bool_array(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> BoolArrayLike: ...


def val_bool_array(
    obj: Any, *, strict: bool = False, cast: bool = True
) -> BoolArrayLike:
    """Validate and optionally cast an object to a boolean array.

    Args:
        obj: The object to validate (array-like or scalar).
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the default boolean dtype.

    Returns:
        The validated boolean array.

    Raises:
        TypeError: If validation fails.

    Examples:
        Validate and cast an object to a boolean array:

        ```python
        import sax.saxtypes as sxt
        import jax.numpy as jnp

        # Valid boolean arrays
        result = sxt.val_bool_array([True, False, True])
        result = sxt.val_bool_array(jnp.array([1, 0, 1]))
        result = sxt.val_bool_array([[True, False], [False, True]])
        ```
    """
    return _val_array_type(
        obj,
        strict=strict,
        cast=cast,
        type_def=BoolArray if strict else BoolArrayLike,
        default_dtype=np.bool_,
        type_name="BoolArray" if strict else "BoolArrayLike",
    )


BoolArray: TypeAlias = Annotated[Array, np.bool_, val(val_bool_array, strict=False)]
"""N-dimensional boolean array (JAX Array with boolean dtype)."""


@overload
def val_int_array(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> IntArray: ...


@overload
def val_int_array(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> IntArrayLike: ...


def val_int_array(obj: Any, *, strict: bool = False, cast: bool = True) -> IntArrayLike:
    """Validate and optionally cast an object to an integer array.

    Args:
        obj: The object to validate (array-like or scalar).
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the default integer dtype.

    Returns:
        The validated integer array.

    Raises:
        TypeError: If validation fails.

    Examples:
        Validate and cast an object to an integer array:

        ```python
        import sax.saxtypes as sxt
        import jax.numpy as jnp

        # Valid integer arrays
        result = sxt.val_int_array([1, 2, 3])
        result = sxt.val_int_array(jnp.array([1.0, 2.0, 3.0]))  # cast from float
        result = sxt.val_int_array([[1, 2], [3, 4]])
        ```
    """
    return _val_array_type(
        obj,
        strict=strict,
        cast=cast,
        type_def=IntArray if strict else IntArrayLike,
        default_dtype=np.int64 if _x64_enabled() else np.int32,
        type_name="IntArray" if strict else "IntArrayLike",
    )


IntArray: TypeAlias = Annotated[
    Array, np.signedinteger, val(val_int_array, strict=False)
]
"""N-dimensional signed integer array (JAX Array with integer dtype)."""


@overload
def val_float_array(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> FloatArray: ...


@overload
def val_float_array(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> FloatArrayLike: ...


def val_float_array(
    obj: Any, *, strict: bool = False, cast: bool = True
) -> FloatArrayLike:
    """Validate and optionally cast an object to a float array.

    Args:
        obj: The object to validate (array-like or scalar).
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the default float dtype.

    Returns:
        The validated float array.

    Raises:
        TypeError: If validation fails.

    Examples:
        Validate and cast an object to a float array:

        ```python
        import sax.saxtypes as sxt
        import jax.numpy as jnp

        # Valid float arrays
        result = sxt.val_float_array([1.0, 2.5, 3.14])
        result = sxt.val_float_array([1, 2, 3])  # cast from int
        result = sxt.val_float_array(jnp.array([[1.0, 2.0], [3.0, 4.0]]))
        ```
    """
    return _val_array_type(
        obj,
        strict=strict,
        cast=cast,
        type_def=FloatArray if strict else FloatArrayLike,
        default_dtype=np.float64 if _x64_enabled() else np.float32,
        type_name="FloatArray" if strict else "FloatArrayLike",
    )


FloatArray: TypeAlias = Annotated[
    Array, np.floating, val(val_float_array, strict=False)
]
"""N-dimensional floating-point array (JAX Array with float dtype)."""


@overload
def val_complex_array(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> ComplexArray: ...


@overload
def val_complex_array(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> ComplexArrayLike: ...


def val_complex_array(
    obj: Any, *, strict: bool = False, cast: bool = True
) -> ComplexArrayLike:
    """Validate and optionally cast an object to a complex array.

    Args:
        obj: The object to validate (array-like or scalar).
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the default complex dtype.

    Returns:
        The validated complex array.

    Raises:
        TypeError: If validation fails.

    Examples:
        Validate and cast an object to a complex array:

        ```python
        import sax.saxtypes as sxt
        import jax.numpy as jnp

        # Valid complex arrays
        result = sxt.val_complex_array([1 + 2j, 3 + 4j])
        result = sxt.val_complex_array([1.0, 2.0, 3.0])  # cast from float
        result = sxt.val_complex_array(jnp.array([[1, 2], [3, 4]]))  # cast from int
        ```
    """
    return _val_array_type(
        obj,
        strict=strict,
        cast=cast,
        type_def=ComplexArray if strict else ComplexArrayLike,
        default_dtype=np.complex128 if _x64_enabled() else np.complex64,
        type_name="ComplexArray" if strict else "ComplexArrayLike",
    )


ComplexArray: TypeAlias = Annotated[
    Array, np.complexfloating, val(val_complex_array, strict=False)
]
"""N-dimensional complex array (JAX Array with complex dtype)."""


@overload
def val_int_array_1d(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> IntArray1D: ...


@overload
def val_int_array_1d(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> IntArray1DLike: ...


def val_int_array_1d(
    obj: Any, *, strict: bool = False, cast: bool = True
) -> IntArray1DLike:
    """Validate and optionally cast an object to a 1D integer array.

    Args:
        obj: The object to validate (1D array-like or scalar).
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the default integer dtype.

    Returns:
        The validated 1D integer array.

    Raises:
        TypeError: If validation fails or array is not 1D.

    Examples:
        Validate and cast an object to a 1D integer array:

        ```python
        import sax.saxtypes as sxt

        # Valid 1D integer arrays
        result = sxt.val_int_array_1d([1, 2, 3, 4])
        result = sxt.val_int_array_1d([1.0, 2.0, 3.0])  # cast from float
        ```
    """
    return _val_array_type(
        obj,
        strict=strict,
        cast=cast,
        type_def=IntArray1D if strict else IntArray1DLike,
        default_dtype=np.int64 if _x64_enabled() else np.int32,
        type_name="IntArray1D" if strict else "Intarray1DLike",
    )


IntArray1D: TypeAlias = Annotated[
    ArrayLike, np.signedinteger, 1, val(val_int_array_1d, strict=False)
]
"""1-dimensional signed integer array (JAX Array with integer dtype)."""


@overload
def val_float_array_1d(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> FloatArray1D: ...


@overload
def val_float_array_1d(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> FloatArray1DLike: ...


def val_float_array_1d(
    obj: Any, *, strict: bool = False, cast: bool = True
) -> FloatArray1DLike:
    """Validate and optionally cast an object to a 1D float array.

    Args:
        obj: The object to validate (1D array-like or scalar).
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the default float dtype.

    Returns:
        The validated 1D float array.

    Raises:
        TypeError: If validation fails or array is not 1D.

    Examples:
        Validate and cast an object to a 1D float array:

        ```python
        import sax.saxtypes as sxt

        # Valid 1D float arrays
        result = sxt.val_float_array_1d([1.0, 2.5, 3.14])
        result = sxt.val_float_array_1d([1, 2, 3])  # cast from int
        ```
    """
    return _val_array_type(
        obj,
        strict=strict,
        cast=cast,
        type_def=FloatArray1D if strict else FloatArray1DLike,
        default_dtype=np.float64 if _x64_enabled() else np.float32,
        type_name="FloatArray1D" if strict else "FloatArray1DLike",
    )


FloatArray1D: TypeAlias = Annotated[
    ArrayLike, np.floating, 1, val(val_float_array_1d, strict=False)
]
"""1-dimensional floating-point array (JAX Array with float dtype)."""


@overload
def val_float_array_2d(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> FloatArray2D: ...


@overload
def val_float_array_2d(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> FloatArray2DLike: ...


def val_float_array_2d(
    obj: Any, *, strict: bool = False, cast: bool = True
) -> FloatArray2DLike:
    """Validate and optionally cast an object to a 2D float array.

    Args:
        obj: The object to validate (2D array-like).
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the default float dtype.

    Returns:
        The validated 2D float array.

    Raises:
        TypeError: If validation fails or array is not 2D.

    Examples:
        Validate and cast an object to a 2D float array:

        ```python
        import sax.saxtypes as sxt

        # Valid 2D float arrays
        result = sxt.val_float_array_2d([[1.0, 2.0], [3.0, 4.0]])
        result = sxt.val_float_array_2d([[1, 2], [3, 4]])  # cast from int
        ```
    """
    return _val_array_type(
        obj,
        strict=strict,
        cast=cast,
        type_def=FloatArray2D if strict else FloatArray2DLike,
        default_dtype=np.float64 if _x64_enabled() else np.float32,
        type_name="FloatArray2D" if strict else "FloatArray2DLike",
    )


FloatArray2D: TypeAlias = Annotated[
    ArrayLike, np.floating, 2, val(val_float_array_2d, strict=False)
]
"""2-dimensional floating-point array (JAX Array with float dtype)."""


@overload
def val_complex_array_1d(
    obj: Any, *, strict: bool = ..., cast: Literal[True] = True
) -> ComplexArray1D: ...


@overload
def val_complex_array_1d(
    obj: Any, *, strict: bool = ..., cast: Literal[False] = False
) -> ComplexArray1DLike: ...


def val_complex_array_1d(
    obj: Any, *, strict: bool = False, cast: bool = True
) -> ComplexArray1DLike:
    """Validate and optionally cast an object to a 1D complex array.

    Args:
        obj: The object to validate (1D array-like or scalar).
        strict: Whether to use strict validation.
        cast: Whether to cast the result to the default complex dtype.

    Returns:
        The validated 1D complex array.

    Raises:
        TypeError: If validation fails or array is not 1D.

    Examples:
        Validate and cast an object to a 1D complex array:

        ```python
        import sax.saxtypes as sxt

        # Valid 1D complex arrays
        result = sxt.val_complex_array_1d([1 + 2j, 3 + 4j])
        result = sxt.val_complex_array_1d([1.0, 2.0, 3.0])  # cast from float
        result = sxt.val_complex_array_1d([1, 2, 3])  # cast from int
        ```
    """
    return _val_array_type(
        obj,
        strict=strict,
        cast=cast,
        type_def=ComplexArray1D if strict else ComplexArray1DLike,
        default_dtype=np.complex128 if _x64_enabled() else np.complex64,
        type_name="ComplexArray1D" if strict else "ComplexArray1DLike",
    )


ComplexArray1D: TypeAlias = Annotated[
    ArrayLike, np.complexfloating, 1, val(val_complex_array_1d, strict=False)
]
"""1-dimensional complex array (JAX Array with complex dtype)."""

BoolLike: TypeAlias = Annotated[bool | np.bool_, val(val_bool, cast=False)]
"""Anything that can be cast into a Bool without loss of data."""

IntLike: TypeAlias = Annotated[int | np.integer, val(val_int, cast=False)]
"""Anything that can be cast into an Int without loss of data."""

FloatLike: TypeAlias = Annotated[
    IntLike | float | np.floating, val(val_float, cast=False)
]
"""Anything that can be cast into a Float without loss of data."""

ComplexLike: TypeAlias = Annotated[
    FloatLike | complex | np.inexact, val(val_complex, cast=False)
]
"""Anything that can be cast into a Complex without loss of data."""

BoolArrayLike: TypeAlias = Annotated[
    ArrayLike | BoolLike, np.bool_, val(val_bool_array, cast=False)
]
"""Anything that can be cast into an N-dimensional Bool array without loss of data."""

IntArrayLike: TypeAlias = Annotated[
    ArrayLike | IntLike, np.integer, val(val_int_array, cast=False)
]
"""Anything that can be cast into an N-dimensional Int array without loss of data."""

FloatArrayLike: TypeAlias = Annotated[
    ArrayLike | FloatLike, np.floating, val(val_float_array, cast=False)
]
"""Anything that can be cast into an N-dimensional Float array without loss of data."""

ComplexArrayLike: TypeAlias = Annotated[
    ArrayLike | ComplexLike, np.inexact, val(val_complex_array, cast=False)
]
"""Anything that can be cast into an N-dim Complex array without loss of data."""

IntArray1DLike: TypeAlias = Annotated[
    IntArrayLike, np.integer, 1, val(val_int_array_1d, cast=False)
]
"""Anything that can be cast into a 1-dimensional integer array without loss of data."""

FloatArray1DLike: TypeAlias = Annotated[
    FloatArrayLike, np.floating, 1, val(val_float_array_1d, cast=False)
]
"""Anything that can be cast into a 1-dimensional float array without loss of data."""

FloatArray2DLike: TypeAlias = Annotated[
    FloatArrayLike, np.floating, 2, val(val_float_array_2d, cast=False)
]
"""Anything that can be cast into a 2-dimensional float array without loss of data."""

ComplexArray1DLike: TypeAlias = Annotated[
    ComplexArrayLike, np.inexact, 1, val(val_complex_array_1d, cast=False)
]
"""Anything that can be cast into a 1-dimensional complex array without loss of data."""


def cast_string(obj: Any) -> str:
    """Cast an object to a string, decoding bytes if necessary.

    Args:
        obj: The object to cast to string.

    Returns:
        The string representation of the object.

    Examples:
        Validate and cast various types to string:

        ```python
        import sax.saxtypes.core as core

        result = core.cast_string("hello")  # "hello"
        result = core.cast_string(b"hello")  # "hello" (decoded)
        result = core.cast_string(42)  # "42"
        ```
    """
    if isinstance(obj, bytes):
        obj = obj.decode()
    return str(obj)


def val_name(
    obj: Any, *, type_name: str = "Name", extra_allowed_chars: tuple[str, ...] = ()
) -> str:
    """Validate that a string is a valid Python identifier.

    Args:
        obj: The object to validate as a name.
        type_name: The type name for error messages.
        extra_allowed_chars: Additional characters to allow beyond standard identifiers.

    Returns:
        The validated name string.

    Raises:
        TypeError: If the string is not a valid identifier.

    Examples:
        Validate a string as a valid Python identifier:

        ```python
        import sax.saxtypes.core as core

        # Valid names
        result = core.val_name("my_var")  # "my_var"
        result = core.val_name("Component1")  # "Component1"

        # With extra allowed characters
        result = core.val_name("inst.port", extra_allowed_chars=(".",))  # "inst.port"
        ```
    """
    s = cast_string(obj)
    _s = s
    for c in extra_allowed_chars:
        _s = _s.replace(c, "_")
    if not _s.isidentifier():
        msg = (
            f"A {type_name!r} string should be a valid python identifier. Got: {s!r}. "
            "note: python identifiers should only contain letters, numbers or "
            "underscores. The first character should not be a number."
        )
        raise TypeError(msg)
    return s


Name: TypeAlias = Annotated[str, val(val_name)]
"""A valid Python identifier - contains only letters, numbers, and underscores."""


class IOLike(Protocol):
    """Protocol for file-like objects that can be read from.

    This protocol defines the interface for objects that can be used
    for reading text data, such as file objects or StringIO objects.
    """

    def read(self, size: int | None = -1, /) -> str:
        """Read and return up to size characters.

        Args:
            size: Maximum number of characters to read. If None or -1, read all.

        Returns:
            The read characters as a string.
        """
        ...


def _get_annotated_type(annotated: Any) -> type | UnionType:
    """Extract the base type from an Annotated type alias."""
    return get_args(annotated)[0]


def _get_annotated_dtype(annotated: Any) -> Any:
    """Extract the dtype from an Annotated array type alias."""
    return get_args(annotated)[1]


def _get_annotated_ndim(annotated: Any) -> Any:
    """Extract the number of dimensions from an Annotated array type alias."""
    with suppress(Exception):
        return int(get_args(annotated)[2])
    return None


def _val_array(obj: Any, error_message: str = "") -> Array:
    """Convert an object to a JAX array with error handling.

    Args:
        obj: The object to convert to an array.
        error_message: Additional error message context.

    Returns:
        The JAX array representation of the object.

    Raises:
        TypeError: If the object cannot be converted to an array.
    """
    try:
        return jnp.asarray(obj)
    except TypeError as e:
        msg = f"NOT_ARRAYLIKE: {error_message}."
        raise TypeError(msg) from e


def _val_0d(obj: Any, *, type_name: str = "0D") -> Array:
    """Validate that an object can be converted to a 0D (scalar) array.

    Args:
        obj: The object to validate.
        type_name: The type name for error messages.

    Returns:
        A 0D JAX array containing the scalar value.

    Raises:
        TypeError: If the object has more than 0 dimensions.
    """
    short_message = f"Cannot cast {obj!r} to {type_name}"
    arr = _val_array(obj)
    if arr.ndim > 0:
        msg = f"NOT_{type_name.upper()}: {short_message}. Got {arr.ndim}D."
        raise TypeError(msg)
    return arr


def _x64_enabled() -> bool:
    """Check if JAX is configured to use 64-bit precision.

    Returns:
        True if 64-bit precision is enabled, False otherwise.
    """
    return jax.config.jax_enable_x64  # type: ignore[reportAttributeAccessIssue]
